"""
Optimizer objects live here.
"""

import os
from inspyred import ec
from inspyred.ec import observers
from inspyred.ec import terminators
from inspyred.ec import selectors
from inspyred.ec import replacers
from inspyred.ec import variators
from random import Random
from time import time
import logging


class __Optimizer(object):
    """
    Base optimization class
    """

    def __init__(
        self,
        max_constraints,
        min_constraints,
        evaluator,
        mutation_rate,
        maximize,
        seeds,
        population_size,
    ):

        self.max_constraints = max_constraints
        self.min_constraints = min_constraints
        self.evaluator = evaluator
        self.population_size = population_size
        self.maximize = maximize
        self.mutation_rate = mutation_rate
        self.seeds = seeds

        # check that constraints are of equal lengths
        assert (len(self.max_constraints) == len(self.min_constraints)), "Min ({}) and max ({}) constraint lists are not of equal size.".format(len(self.min_constraints), len(self.max_constraints))

        # check that min constraints are always less than max constraints
        for i in range(0, len(self.max_constraints)):
            assert (self.min_constraints[i] <= self.max_constraints[i]), "Min constraint {} is larger than max constraint {}.".format(self.min_constraints[i], self.max_constraints[i])

    def uniform_random_chromosome(self, random, args):
        chromosome = []
        for lo, hi in zip(self.max_constraints, self.min_constraints):
            chromosome.append(random.uniform(lo, hi))
        return chromosome

    def print_report(self, final_pop, do_plot, stat_file_name):
        print(max(final_pop))
        # Sort and print the fittest individual, which will be at index 0.
        final_pop.sort(reverse=True)
        print("\n  Fittest individual:\n  %s" % final_pop[0])

        if do_plot:
            from inspyred.ec import analysis

            try:
                stat_file = open(stat_file_name, "r")
                analysis.generation_plot(stat_file, errorbars=False)
            finally:
                stat_file.close()


class CustomOptimizerA(__Optimizer):
    def __init__(
        self,
        max_constraints,
        min_constraints,
        evaluator,
        mutation_rate=0.2,
        max_evaluations=100,
        population_size=10,
        num_selected=None,
        tourn_size=2,
        num_elites=1,
        maximize=False,
        num_offspring=None,
        seeds=[],
        verbose=False,
    ):

        super(CustomOptimizerA, self).__init__(
            max_constraints,
            min_constraints,
            evaluator,
            mutation_rate,
            maximize,
            seeds,
            population_size,
        )

        self.max_evaluations = max_evaluations
        self.tourn_size = tourn_size
        self.num_elites = num_elites
        self.mutation_rate = mutation_rate
        self.verbose = verbose

        if num_selected == None:
            self.num_selected = population_size
        else:
            self.num_selected = num_selected
        if num_offspring == None:
            self.num_offspring = population_size - self.num_selected
        else:
            self.num_offspring = num_offspring

    def optimize(self, do_plot=True, seed=int(time()), summary_dir=None):

        rand = Random()
        rand.seed(seed)

        if summary_dir is None:
            cwd = os.getcwd()
            summary_dir = os.path.dirname(cwd) + "/data/"

        if not os.path.exists(summary_dir):
            os.mkdir(summary_dir)

        stat_file_name = summary_dir + "/ga_statistics.csv"
        ind_file_name = summary_dir + "/ga_individuals.csv"

        stat_file = open(stat_file_name, "w")
        ind_file = open(ind_file_name, "w")
        print("Created files: %s and %s" % (stat_file_name, ind_file_name))

        if self.verbose:
            logger = logging.getLogger("inspyred.ec")
            logger.setLevel(logging.DEBUG)

            ch = logging.StreamHandler()
            # ch.setLevel(logging.DEBUG)
            formatter = logging.Formatter(">>> EC: - %(levelname)s - %(message)s")
            ch.setFormatter(formatter)
            logger.addHandler(ch)

        algorithm = ec.EvolutionaryComputation(rand)
        algorithm.observer = observers.file_observer
        algorithm.terminator = terminators.evaluation_termination
        algorithm.selector = selectors.tournament_selection
        algorithm.replacer = replacers.steady_state_replacement
        algorithm.variator = [variators.blend_crossover, variators.gaussian_mutation]

        final_pop = algorithm.evolve(
            generator=self.uniform_random_chromosome,
            evaluator=self.evaluator.evaluate,
            pop_size=self.population_size,
            maximize=self.maximize,
            bounder=ec.Bounder(
                lower_bound=self.min_constraints, upper_bound=self.max_constraints
            ),
            num_selected=self.num_selected,
            tourn_size=self.tourn_size,
            num_elites=self.num_elites,
            num_offspring=self.num_offspring,
            max_evaluations=self.max_evaluations,
            mutation_rate=self.mutation_rate,
            statistics_file=stat_file,
            seeds=self.seeds,
            individuals_file=ind_file,
        )

        stat_file.close()
        ind_file.close()

        self.print_report(final_pop, do_plot, stat_file_name)

        # return the parameter set for the best individual

        return final_pop[0].candidate, final_pop[0].fitness



"""
The controllers module provides different controller classes,
applicable to different simulations.

A controller object's job is to control simulations-
At a high level a controller objects  accepts a list of
parameters and chromosomes and (usually) returns
corresponding simulation data.
This is implemented polymporphically in subclasses.
Each controller class must therefore provide a run method, which is used by
the evaluator to run a simulation.

A controller must be able to accept simulation parameters (chromosomes)
from the evaluator.

The evaluator is therefore only concerned with assigining fitness to chromosomes.

On the whole this allows for deep modularization -
as long as the user can provide a controller which will (for instance)
reutrn sample and time arrays for arbitrary chromosome and parameter
lists a range of evaluators would be able to utilise it.
"""

import os
import subprocess

import math


class __Controller:
    """
    Controller base class
    """

    def run(self, candidates, parameters):
        """
        At a high level - accepts a list of parameters and chromosomes
        and (usually) returns corresponding simulation data. This is
        implemented polymporphically in subclasses.
        """
        raise NotImplementedError("Valid controller requires run method!")


class CLIController(__Controller):
    """
    Control simulations via command line arguments executed through the Python os module.
    """

    def __init__(self, cli_argument):
        self.cli_argument = cli_argument

    def run(self, candidates, parameters, fitness_filename="evaluations"):

        # "Run simulation"

        for chromosome in candidates:
            self.chromosome = chromosome
            self.parameters = parameters  # actually unneeded
            # this manipulation is slightly messy, done for conversion of chromosome
            # into something that can be executed on the shell
            chromosome_str = "".join(str(e) + " " for e in chromosome)
            cla = self.cli_argument + " " + fitness_filename + " " + chromosome_str
            print(cla)
            subprocess.call(cla, shell=True)


class NrnProject(__Controller):
    """
    Run an nrnproject simulation based on optimizer parameters."""

    def __init__(self, nrnproject_path, db_path, exp_id=None):

        self.sim_main_path = os.path.join(nrnproject_path, "src/simrunner.py")
        self.nrnproject_path = nrnproject_path
        self.db_path = db_path
        self.exp_id = exp_id

    def __generate_cla(self):
        sim_var_string = self.__generate_sim_var_string()
        cla = "python " + self.sim_main_path + sim_var_string
        return cla

    def __generate_sim_var_string(self):
        sim_var_string = ""
        for i in enumerate(self.parameters):
            sim_var_string += (
                " \"sim_var['" + i[1] + "'] = " + str(self.chromosome[i[0]]) + '"'
            )

        if self.exp_id != None:
            sim_var_string += " \"sim_var['exp_id'] =" + str(self.exp_id) + '"'

        return sim_var_string

    def run(self, candidates, parameters):

        # """Run simulations"""

        import sqldbutils

        exp_data_array = []
        for chromosome in candidates:
            self.chromosome = chromosome
            self.parameters = parameters
            exp_id = sqldbutils.generate_exp_ids(self.db_path)
            cla = self.__generate_cla()
            os.chdir(self.nrnproject_path + "/src/")  # there should be a smarter way
            os.system(cla)
            print(self.db_path)
            print(exp_id)
            exp_data = sqldbutils.sim_data(self.db_path, exp_id)
            exp_data_array.append(exp_data)
        return exp_data_array


class __CondorContext(object):
    """Context for Condor-based grid"""

    def __init__(self, host, username, password, port):

        self.messagehost = ssh_utils.host(host, username, password, port)

    def __split_list(self, alist, wanted_parts=1):

        length = len(alist)
        return [
            alist[i * length // wanted_parts : (i + 1) * length // wanted_parts]
            for i in range(wanted_parts)
        ]

    def __prepare_candidates(self, candidates, candidates_per_job=1):

        # Split candidate list into smaller ones (jobs):
        # and make a job list
        if optimizer_params.candidates_in_job != None:
            candidates_in_job = optimizer_params.candidates_in_job
        else:
            candidates_in_job = candidates_per_job

        num_candidates = len(candidates)
        ids = range(num_candidates)
        enumerated_candidates = zip(candidates, ids)
        num_jobs = num_candidates / candidates_in_job
        self.num_jobs = num_jobs
        self.job_list = self.__split_list(
            enumerated_candidates, wanted_parts=self.num_jobs
        )

    def __make_job_file(self, job, job_number):
        # write the header:
        filepath = os.path.join(self.tmpdir, "run" + str(job_number) + ".sh")
        run_shell = open(filepath, "w")
        run_shell.write("#!/bin/bash\n")
        run_shell.write("reldir=`dirname $0`\n")
        run_shell.write("cd $reldir\n")
        run_shell.write("directory=`pwd`\n")
        run_shell.write("pndirectory=$directory\n")
        run_shell.write("#Untar the file:\n")
        run_shell.write("/bin/tar xzf ./portable-neuron.tar.gz\n")
        tarfile_name = optimizer_params.tarred_nrnproj
        run_shell.write("/bin/tar xzf ./" + tarfile_name + "\n")

        # CandidateData_list=[]
        for enumerated_candidate in job:

            chromosome = enumerated_candidate[0]
            candidate_info = CandidateData(chromosome)
            exp_id = enumerated_candidate[1]
            candidate_info.set_exp_id(exp_id)
            candidate_info.set_job_num(job_number)
            self.CandidateData_list.append(candidate_info)

            nproj = controllers.NrnProjSimRun(optimizer_params.project_path, chromosome)
            run_shell.write("#issue the commands\n")
            run_shell.write(
                "$pndirectory/pnpython.sh \
                    $directory/src/simrunner.py \"sim_var['exp_id'] \
                    = "
                + str(exp_id)
                + '" '
                + "\"sim_var['"
                "dbname"
                "'] \
                    = 'outputdb"
                + str(job_number)
                + ".sqlite'\""
                + nproj.sim_var_string
                + "\n"
            )
            run_shell.write("echo 'done'\n")

        run_shell.write(
            "cp $directory/sims/outputdb" + str(job_number) + ".sqlite $directory\n"
        )
        # self.CandidateData_list=CandidateData_list
        run_shell.close()

    def __make_submit_file(self):
        """write the condor submit files"""

        filepath = os.path.join(self.tmpdir, "submitfile.submit")
        submit_file = open(filepath, "w")

        submit_file.write("universe   = vanilla\n")
        submit_file.write("log        = pneuron.log\n")
        submit_file.write("Error   = err.$(Process)\n")
        submit_file.write("Output  = out.$(Process)\n")
        submit_file.write('requirements = GLIBC == "2.11"\n')
        tarfile_name = optimizer_params.tarred_nrnproj
        submit_file.write(
            "transfer_input_files = portable-neuron.tar.gz," + tarfile_name + "\n"
        )
        submit_file.write("should_transfer_files   = yes\n")
        submit_file.write("when_to_transfer_output = on_exit_or_evict\n")
        # this is where you have to do the clever stuff:

        for shellno in range(self.num_jobs):
            submit_file.write("executable = run" + str(shellno) + ".sh\n")
            submit_file.write("queue\n")

        # finally close the submit file
        submit_file.close()

    def __build_condor_files(self, candidates, parameters, candidates_per_job=100):

        # prepare list of candidates to be farmed on grid:
        self.__prepare_candidates(candidates, candidates_per_job=100)

        # make the job files (shell scripts to be executed on the execute nodes)
        job_number = 0  # run shell script number
        for job in self.job_list:
            self.__make_job_file(job, job_number)
            job_number += 1

        # now make the submit file
        self.__make_submit_file()

    def __delete_remote_files(self, host):
        import ssh_utils

        command = "rm -rf ./*"
        ssh_utils.issue_command(host, command)

    def __put_multiple_files(self, host, filelist, localdir="/", remotedir="/"):
        import ssh_utils

        for file in filelist:
            localpath = os.path.join(localdir, file)
            remotepath = os.path.join(remotedir, file)
            ssh_utils.put_file(host, localpath, remotepath)


class NrnProjectCondor(NrnProject):
    """
    Run NrnProject-based simulations on a Condor-managed
    federated system
    """

    def __init__(
        self,
        host,
        username,
        password,
        port=80,
        local_analysis=False,
        candidates_per_job=100,
    ):

        super(NrnProjectCondor, self).__init__()

        # other things like the number of nodes to divide the work onto and
        # host connection parameters need to go into this constructor
        # the more I think about it the less this seems like a good idea
        # though

        if local_analysis:
            self.run = self.__local_run
        else:
            self.run = self.__remote_run__

        # make a context which provides grid utilities
        self.context = __CondorContext(host, username, password, port)

        self.cpj = candidates_per_job

    def __condor_run(self, candidates, parameters):
        """
        Run simulations on grid and analyse data locally (???I'm quite confused here...there is a mistake somewhere as the name doesn't match the description - which method is which?)

            Once each generation has finished, all data is pulled to local
            workstation in form of sqlite databases (1 database per job)
            and these are analysed and the fitness estimated sequentially
            the fitness array is then returned.
        """

        import time
        import ssh_utils

        # Build submit and runx.sh files, exp_id now corresponds
        # to position in chromosome and fitness arrays
        self.context.__build_condor_files(
            candidates, parameters, candidates_per_job=self.cpj
        )

        # This is a file handling block..
        # delete everything in the ssh_utilse directory you're about to put files in
        self.__delete_remote_files__()
        filelist = os.listdir(self.tmpdir)
        # copy local files over, some stuff is missing here as it needs to be an attribute in the condor context
        self.__put_multiple_files(filelist, localdir=self.tmpdir)
        filelist = os.listdir(self.portableswdir)
        # copy portable software files over:
        self.__put_multiple_files(filelist, localdir=self.portableswdir)

        # issue a command to the message host to issue commands to the grid:
        ssh_utils.issue_command(
            context.messagehost,
            "export PATH=/opt/Condor/release/bin:$PATH\ncondor_submit submitfile.submit",
        )

        # make a list of the database files we need:
        self.jobdbnames = []
        for job_num in range(self.num_jobs):
            jobdbname = "outputdb" + str(job_num) + ".sqlite"
            self.jobdbnames.append(jobdbname)

        # wait till we know file exists:
        dbs_created = False
        pulled_dbs = (
            []
        )  # list of databases which have been extracted from remote server
        while dbs_created == False:
            print("waiting..")
            time.sleep(20)
            print("checking if dbs created:")
            command = "ls"
            remote_filelist = ssh_utils.issue_command(self.messagehost, command)
            for jobdbname in self.jobdbnames:
                db_exists = jobdbname + "\n" in remote_filelist
                if db_exists == False:
                    print(jobdbname + " has not been generated")
                    dbs_created = False
                elif db_exists == True and jobdbname not in pulled_dbs:
                    print(jobdbname + " has been generated")
                    remotefile = optimizer_params.remotedir + jobdbname
                    localpath = os.path.join(
                        self.datadir, str(self.generation) + jobdbname
                    )
                    ssh_utils.get_file(self.messagehost, remotefile, localpath)
                    pulled_dbs.append(
                        jobdbname
                    )  # so that it is not extracted more than once
                    # here pop-in the fitness evaluation
                if len(pulled_dbs) == len(self.jobdbnames):
                    dbs_created = True

        # this block can be simplified, it need simply return exp_data containers
        fitness = []
        for CandidateData in self.CandidateData_list:
            job_num = CandidateData.job_num

            dbname = str(self.generation) + "outputdb" + str(job_num) + ".sqlite"
            dbpath = os.path.join(self.datadir, dbname)
            exp_id = CandidateData.exp_id

            connection = sqldbutils.db_connect(
                dbpath
            )  # establish a database connection
            query = (
                "SELECT numerical_value\
                    FROM output_params WHERE experiment_id=\
                    "
                + str(exp_id)
                + ' AND parameter="fitness"'
            )

            exp_fitness = sqldbutils.execute_query(connection, query)
            exp_fitness = exp_fitness.fetchall()
            exp_fitness = exp_fitness[0][0]

            print("Fitness:")
            print(exp_fitness)

            fitness.append(exp_fitness)

        self.generation += 1
        return fitness

    ###ignore this for now###
    def __local_evaluate(self, candidates, args):
        import time

        analysis

        self.CandidateData_list = []
        analysis_var = self.analysis_var

        # Build submitfile.submit and runx.sh files:
        self.__buil_condor_files(
            candidates
        )  # exp_id now corresponds to position in chromosome/fitness array

        fitness = []
        # submit the jobs to the grid
        os.chdir(self.tmpdir)
        os.system("condor_submit submitfile.submit")

        # wait till you know file exists:
        dbs_created = False
        while dbs_created == False:
            print("checking if dbs created:")
            for job_num in range(self.num_jobs):
                jobdbname = "outputdb" + str(job_num) + ".sqlite"
                jobdbpath = os.path.join(self.datadir, jobdbname)
                print(jobdbpath)
                db_exists = os.path.exists(jobdbpath)

                if db_exists == False:
                    time.sleep(60)
                    dbs_created = False
                    break

                dbs_created = True

        for CandidateData in self.CandidateData_list:
            job_num = CandidateData.job_num
            dbname = "/outputdb" + str(job_num) + ".sqlite"
            dbpath = self.datadir + dbname
            exp_id = CandidateData.exp_id
            exp_data = sqldbutils.sim_data(dbpath, exp_id)
            analysis = analysis.IClampAnalysis(
                exp_data.samples, exp_data.t, analysis_var, 5000, 10000
            )
            exp_fitness = analysis.evaluate_fitness(
                self.targets,
                self.weights,
                cost_function=analysis.normalised_cost_function,
            )
            fitness.append(exp_fitness)

        for job_num in range(self.num_jobs):
            jobdbname = "outputdb" + str(job_num) + ".sqlite"
            jobdbpath = os.path.join(self.datadir, jobdbname)
            print(jobdbpath)
            os.remove(jobdbpath)

        return fitness


class SineWaveController(__Controller):
    """
    Simple sine wave generator which takes a number of variables ('amp', 'period', 'offset')
    and produces an output based on these.
    """

    def __init__(self, sim_time, dt):

        self.sim_time = sim_time
        self.dt = dt

    def run_individual(self, sim_var, gen_plot=False, show_plot=False):
        """
        Run an individual simulation.

        The candidate data has been flattened into the sim_var dict. The
        sim_var dict contains parameter:value key value pairs, which are
        applied to the model before it is simulated.

        """
        print(">> Running individual: %s" % (sim_var))

        import numpy as np

        t = 0
        times = []
        volts = []

        while t <= self.sim_time:
            v = sim_var["offset"] + (
                sim_var["amp"] * (math.sin(2 * math.pi * t / sim_var["period"]))
            )
            times.append(t)
            volts.append(v)
            t += self.dt

        if gen_plot:
            from matplotlib import pyplot as plt

            info = ""
            for key in sim_var.keys():
                info += "%s=%s " % (key, sim_var[key])
            plt.plot(times, volts, label=info)
            plt.legend(
                loc="upper center",
                bbox_to_anchor=(0.5, -0.05),
                fancybox=True,
                shadow=True,
                ncol=1,
            )

            if show_plot:
                plt.show()

        return np.array(times), np.array(volts)

    def run(self, candidates, parameters):
        """
        Run simulation for each candidate

        This run method will loop through each candidate and run the simulation
        corresponding to its parameter values. It will populate an array called
        traces with the resulting voltage traces for the simulation and return it.
        """

        traces = []
        for candidate in candidates:
            sim_var = dict(zip(parameters, candidate))
            t, v = self.run_individual(sim_var)
            traces.append([t, v])

        return traces



import os
import sys
from threading import Thread
from pyelectro import analysis
import numpy
import math

import pprint

pp = pprint.PrettyPrinter(indent=4)


def alpha_normalised_cost_function(value, target, base=10):
    """Fitness of a value-target pair from 0 to 1

    .. WARNING:
        I've found that this cost function is producing some odd behaviour.
        It is best avoided until this is investigated

    For any value/target pair will give a normalised value for
    agreement 1 is complete value-target match and 0 is 0 match.
    A mirrored exponential function is used.
    The fitness is given by the expression :math:`fitness = base^{-x}`

    where:

    .. math::
          x = {\dfrac{(value-target)}{(target + 0.01)^2}}

    :param value: value measured
    :param t: target
    :param base: the value 'base' in the above mathematical expression for x

    :return: fitness - a real number from 0 to 1

    """

    value = float(value)
    target = float(target)

    x = (
        (value - target) / (target + 0.01)
    ) ** 2  # the 0.01 thing is a bit of a hack at the moment.
    fitness = base ** (-x)
    return fitness


def normalised_cost_function(value, target, Q=None):
    """Returns fitness of a value-target pair from 0 to 1

    For any value/target pair will give a normalised value for
    agreement 0 is complete value-target match and 1 is "no" match.

    If no Q is assigned, it is set such that it satisfies the condition
    fitness=0.7 when (target-valu)e=10*target. This is essentially
    empirical and seems to work. Mathematical derivation is on Mike Vella's
    Lab Book 1 p.42 (page dated 15/12/11).

    :param value: value measured
    :param t: target
    :param Q: This is the sharpness of the cost function, higher values correspond
        to a sharper cost function. A high Q-Value may lead an optimizer to a solution
        quickly once it nears the solution.

    :return: fitness value from 0 to 1

    """

    value = float(value)
    target = float(target)

    if Q is None:
        if target != 0:
            Q = 7 / (300 * (target ** 2))
        else:
            Q = 0.023333  # PG: Gives fitness = 0.023333 when value = 1; fitness = 0.7 when value = 10

    try:
        fitness = 1 - 1 / (Q * (target - value) ** 2 + 1)
    except:
        print(
            "Exeption when calculating the fitness function; target: %s; value %s; Q: %s"
            % (target, value, Q)
        )
        fitness = 1

    return fitness


class __CandidateData(object):
    """Container for information about a candidate (chromosome)"""

    def __init__(self, chromosome):
        self.chromosome = chromosome

    def set_dbpath(self, dbpath):
        self.dbpath = dbpath

    def set_exp_id(self, exp_id):
        self.exp_id = exp_id

    def set_job_num(self, job_num):
        self.job_num = job_num


class __Evaluator(object):
    """Base class for Evaluators"""

    def __init__(self, parameters, weights, targets, controller):

        self.parameters = parameters
        self.weights = weights
        self.targets = targets
        self.controller = controller


'''
    PG: Disabling these until they're tested again...
    
class __CondorContext(object):
    """manager for dealing with a condor-based grid"""
        
    def __split_list(self,alist, wanted_parts=1):
        
        length = len(alist)
        return [ alist[i*length // wanted_parts: (i+1)*length // wanted_parts] 
                 for i in range(wanted_parts) ]
    
    def __prepare_candidates(self,candidates,candidates_per_job=1):
        
        #Split candidate list into smaller ones (jobs):
        #and make a job list
        if optimizer_params.candidates_in_job != None:
            candidates_in_job=optimizer_params.candidates_in_job
        else:
            candidates_in_job=candidates_per_job

        num_candidates=len(candidates)
        ids=range(num_candidates)
        enumerated_candidates=zip(candidates,ids)
        num_jobs=num_candidates/candidates_in_job
        self.num_jobs=num_jobs
        self.job_list=self.__split_list(enumerated_candidates,wanted_parts=self.num_jobs)
    
    def __make_job_file(self,job,job_number):
        #write the header:
        filepath = os.path.join(self.tmpdir, 'run' + str(job_number) + '.sh')
        run_shell = open(filepath, 'w')
        run_shell.write('#!/bin/bash\n')
        run_shell.write('reldir=`dirname $0`\n')
        run_shell.write('cd $reldir\n')
        run_shell.write('directory=`pwd`\n')
        run_shell.write('pndirectory=$directory\n')
        run_shell.write('#Untar the file:\n')
        run_shell.write('/bin/tar xzf ./portable-neuron.tar.gz\n')
        tarfile_name=optimizer_params.tarred_nrnproj
        run_shell.write('/bin/tar xzf ./'+tarfile_name+'\n')
        
        #CandidateData_list=[]
        for enumerated_candidate in job:
                
            chromosome = enumerated_candidate[0]
            candidate_info = CandidateData(chromosome)
            exp_id = enumerated_candidate[1]
            candidate_info.set_exp_id(exp_id)
            candidate_info.set_job_num(job_number)
            self.CandidateData_list.append(candidate_info)
                
            nproj = controllers.NrnProjSimRun(optimizer_params.project_path, chromosome)
            run_shell.write('#issue the commands\n')
            run_shell.write('$pndirectory/pnpython.sh \
                    $directory/src/simrunner.py "sim_var[\'exp_id\'] \
                    = ' + str(exp_id) + '\" ' + '"sim_var[\'''dbname''\'] \
                    = \'outputdb' + str(job_number) + '.sqlite\'"' + 
                    nproj.sim_var_string + '\n')
            run_shell.write('echo \'done\'\n')
        
        run_shell.write('cp $directory/sims/outputdb' + str(job_number) + '.sqlite $directory\n')
        #self.CandidateData_list=CandidateData_list
        run_shell.close()

    def __make_submit_file(self):
        #now we write the submit file
        filepath = os.path.join(self.tmpdir, 'submitfile.submit')
        submit_file=open(filepath,'w')
        
        submit_file.write('universe   = vanilla\n')
        submit_file.write('log        = pneuron.log\n')
        submit_file.write('Error   = err.$(Process)\n')                                                
        submit_file.write('Output  = out.$(Process)\n')
        submit_file.write('requirements = GLIBC == "2.11"\n')
        tarfile_name=optimizer_params.tarred_nrnproj
        submit_file.write('transfer_input_files = portable-neuron.tar.gz,'+tarfile_name+'\n')
        submit_file.write('should_transfer_files   = yes\n')
        submit_file.write('when_to_transfer_output = on_exit_or_evict\n')
        #this is where you have to do the clever stuff:
        
        for shellno in range(self.num_jobs):
            submit_file.write('executable = run'+str(shellno)+'.sh\n')
            submit_file.write('queue\n')
        
        #finally close the submit file
        submit_file.close()
    
    def __build_condor_files(self,candidates,candidates_per_job=100):
        
        #prepare list of candidates to be farmed on grid:
        self.__prepare_candidates(candidates,candidates_per_job=100)
        
        #make the job files (shell scripts to be executed on the execute nodes)
        job_number=0 #run shell script number
        for job in self.job_list:
            self.__make_job_file(job,job_number)
            job_number+=1
        
        #now make the submit file
        self.__make_submit_file()

    def __delete_remote_files(self,host):
        import ssh_utils
        command='rm -rf ./*'
        ssh_utils.issue_command(host, command)
    
    def __put_multiple_files(self,host,filelist,localdir='/',remotedir='/'):
        import ssh_utils
        
        for file in filelist:
            localpath=os.path.join(localdir,file)
            remotepath=os.path.join(remotedir,file)
            ssh_utils.put_file(host,localpath,remotepath)
'''


class DumbEvaluator(__Evaluator):
    """
    The simulations themselves report their fitness. The evaluator
    just reads them from a file. Requires the appropriate controller.
    """

    def __init__(self, controller, fitness_filename_prefix, threads_number=1):
        self.controller = controller
        self.fitness_filename_prefix = fitness_filename_prefix
        self.threads_number = threads_number

    def evaluate(self, candidates, args):

        threads_number = int(self.threads_number)
        candidates_per_thread = (len(candidates)) / threads_number
        remainder_candidates = len(candidates) % threads_number
        chunk_begin = 0
        chunk_end = candidates_per_thread

        if remainder_candidates != 0:
            chunk_end += 1

        threads = []

        try:
            for i in range(0, threads_number):
                # if fitness file exists need to destroy it:
                file_name = self.fitness_filename_prefix + str(i)
                if os.path.exists(file_name):
                    os.remove(file_name)

                # run the candidates:
                candidate_section = candidates[chunk_begin:chunk_end]
                threads.append(
                    Thread(
                        target=self.controller.run,
                        args=(
                            candidate_section,
                            args,
                            file_name,
                        ),
                    )
                )
                threads[i].daemon = True
                threads[i].start()

                chunk_begin = chunk_end
                chunk_end += candidates_per_thread
                if i < (remainder_candidates - 1):
                    chunk_end += 1

            fitness = []
            for i in range(0, threads_number):
                # we should let the main thread handle keybord interrupts
                while True:
                    threads[i].join(1)
                    if not threads[i].isAlive():
                        break

                # get their fitness from the file
                file_name = self.fitness_filename_prefix + str(i)
                threads[i].join()
                fitness = fitness + [float(i) for i in open(file_name).readlines()]
                os.remove(file_name)
        except (KeyboardInterrupt, SystemExit):
            sys.exit("Interrupted by ctrl+c\n")

        return fitness


class IClampEvaluator(__Evaluator):
    """
    Locally-evaluates (not using cluster or grid computing) a model.

    The evaluate routine runs the model and returns its fitness value

    """

    def __init__(
        self,
        analysis_start_time,
        controller,
        analysis_end_time,
        target_data_path,
        parameters,
        analysis_var,
        weights,
        targets=None,
        automatic=False,
        verbose=True,
    ):

        super(IClampEvaluator, self).__init__(parameters, weights, targets, controller)

        self.analysis_start_time = analysis_start_time
        self.analysis_end_time = analysis_end_time
        self.target_data_path = target_data_path
        self.analysis_var = analysis_var
        self.verbose = verbose

        print("target data path in evaluator:" + target_data_path)

        if automatic is True:
            t, v_raw = analysis.load_csv_data(target_data_path)
            v = numpy.array(v_raw)

            v_smooth = list(analysis.smooth(v))

            ic_analysis = analysis.IClampAnalysis(
                v_smooth,
                t,
                analysis_var,
                start_analysis=analysis_start_time,
                end_analysis=analysis_end_time,
            )

            ic_analysis.analyse()

            self.targets = ic_analysis.analysis_results

            print("Obtained targets are:")
            print(self.targets)

    def evaluate(self, candidates, args):

        print("\n>>>>>  Evaluating: ")
        for cand in candidates:
            print(">>>>>       %s" % cand)

        simulations_data = self.controller.run(candidates, self.parameters)

        fitness = []

        for data in simulations_data:

            times = data[0]
            samples = data[1]

            data_analysis = analysis.IClampAnalysis(
                samples,
                times,
                self.analysis_var,
                start_analysis=self.analysis_start_time,
                end_analysis=self.analysis_end_time,
                target_data_path=self.target_data_path,
            )

            try:
                data_analysis.analyse()
            except:
                data_analysis.analysable_data = False

            fitness_value = self.evaluate_fitness(
                data_analysis,
                self.targets,
                self.weights,
                cost_function=normalised_cost_function,
            )
            fitness.append(fitness_value)

            print("Fitness: %s\n" % fitness_value)

        return fitness

    def evaluate_fitness(
        self,
        data_analysis,
        target_dict={},
        target_weights=None,
        cost_function=normalised_cost_function,
    ):
        """
        Return the estimated fitness of the data, based on the cost function being used.
            :param data_analysis: IClampAnalysis instance
            :param target_dict: key-value pairs for targets
            :param target_weights: key-value pairs for target weights
            :param cost_function: cost function (callback) to assign individual targets sub-fitness.
        """

        # calculate max fitness value (TODO: there may be a more pythonic way to do this..)
        worst_cumulative_fitness = 0
        for target in target_dict.keys():
            if target_weights is None:
                target_weight = 1
            else:
                if target in target_weights.keys():
                    target_weight = target_weights[target]
                else:
                    target_weight = 1.0

            worst_cumulative_fitness += target_weight

        # if we have 1 or 0 peaks we won't conduct any analysis
        if data_analysis.analysable_data is False:
            print("Data is non-analysable")
            return worst_cumulative_fitness

        else:
            fitness = 0

            for target in target_dict.keys():

                target_value = target_dict[target]
                cost = "?"

                if target_weights is None:
                    target_weight = 1
                else:
                    if target in target_weights.keys():
                        target_weight = target_weights[target]
                    else:
                        target_weight = 1.0
                if target_weight > 0:
                    value = data_analysis.analysis_results[target]
                    # let function pick Q automatically
                    cost = cost_function(value, target_value)
                    inc = target_weight * cost
                    fitness += inc
                    if self.verbose:
                        print(
                            "Target %s (weight %s): target val: %s, actual: %s, cost: %s, fitness inc: %s"
                            % (target, target_weight, target_value, value, cost, inc)
                        )

            return fitness


class NetworkEvaluator(__Evaluator):
    """
    Locally-evaluates (not using cluster or grid computing) a model.

    The evaluate routine runs the model and returns its fitness value

    """

    def __init__(
        self,
        analysis_start_time,
        controller,
        analysis_end_time,
        parameters,
        analysis_var,
        weights,
        targets=None,
    ):

        super(NetworkEvaluator, self).__init__(parameters, weights, targets, controller)

        self.analysis_start_time = analysis_start_time
        self.analysis_end_time = analysis_end_time
        self.analysis_var = analysis_var
        self.targets = targets

    def evaluate(self, candidates, args):

        print("\n>>>>>  Evaluating: ")
        for cand in candidates:
            print(">>>>>       %s" % cand)

        simulations_data = self.controller.run(candidates, self.parameters)

        fitness = []

        for i in range(len(simulations_data)):

            data = simulations_data[i]
            candidate = candidates[i]
            times = data[0]
            volts = data[1]

            data_analysis = analysis.NetworkAnalysis(
                volts,
                times,
                self.analysis_var,
                start_analysis=self.analysis_start_time,
                end_analysis=self.analysis_end_time,
            )

            print(
                "- Evaluating %s from %s -> %s (data %s -> %s)"
                % (
                    candidate,
                    self.analysis_start_time,
                    self.analysis_end_time,
                    times[0],
                    times[-1],
                )
            )

            data_analysis.analyse(self.targets)

            fitness_value = self.evaluate_fitness(
                data_analysis,
                self.targets,
                self.weights,
                cost_function=normalised_cost_function,
            )
            fitness.append(fitness_value)

            print("Fitness: %s\n" % fitness_value)

        return fitness

    def evaluate_fitness(
        self,
        data_analysis,
        target_dict={},
        target_weights=None,
        cost_function=normalised_cost_function,
    ):
        """
        Return the estimated fitness of the data, based on the cost function being used.
            :param data_analysis: NetworkAnalysis instance
            :param target_dict: key-value pairs for targets
            :param target_weights: key-value pairs for target weights
            :param cost_function: cost function (callback) to assign individual targets sub-fitness.
        """

        fitness = 0

        for target in target_dict.keys():

            target_value = target_dict[target]

            if target_weights is None:
                target_weight = 1
            else:
                if target in target_weights.keys():
                    target_weight = target_weights[target]
                else:
                    target_weight = 0  # If it's not mentioned assunme weight = 0!

            if target_weight > 0:
                inc = target_weight  # default...
                cost = "?"
                if target in data_analysis.analysis_results:
                    value = data_analysis.analysis_results[target]
                    if not math.isnan(value):
                        # let function pick Q automatically
                        cost = cost_function(value, target_value)
                        inc = target_weight * cost
                    else:
                        value = "<<infinite value!>>"
                        inc = target_weight
                else:
                    # Check if any targets for the provided entity are included
                    # in the analysis results. If not, something is wrong.
                    valid = False
                    entity = target.split("/")[0].split(":")[0]
                    for atarget in data_analysis.analysis_results.keys():
                        if entity in atarget:
                            valid = True
                            break

                    if not valid:
                        raise RuntimeError("No target values for entity {} were found in the analysis results.  Please check your target parameter strings.\nAll target parameters are: {}".format(entity, data_analysis.analysis_results.keys()))

                    value = (
                        "<<cannot be calculated! (only: %s; peak_threshold: %s)>>"
                        % (
                            data_analysis.analysis_results.keys(),
                            self.analysis_var["peak_threshold"],
                        )
                    )

                fitness += inc

                print(
                    "Target %s (weight %s): target val: %s, actual: %s, cost: %s, fitness inc: %s"
                    % (target, target_weight, target_value, value, cost, inc)
                )

        return fitness


'''
class IClampCondorEvaluator(IClampEvaluator):
    """
    Evaluate simulations and return their fitness on a condor grid.
    
        Tested and known to work on CamGrid 
        (http://www.escience.cam.ac.uk/projects/camgrid/)
      
        WARNING:
        this entire class should now be considered obsolete, the evaluator
        is just an IClampEvaluator and everything here that is different
        from that class needs to become its own controller

    """

    def __init__(self,local_analysis=False):

        super(IClampCondorEvaluator,self).__init__()

        #other things like the number of nodes to divide the work onto and
        #host connection parameters need to go into this constructor

        if local_analysis:
            self.evaluate=self.__local_evaluate
        else:
            self.evaluate=self.__remote_evaluate__

    def __condor_evaluate(self,candidates,args):
        """
        Run simulations on grid and analyse data locally
        WARNING: (???I'm quite confused here...there is a mistake somewhere
        as the name doesn't match the description - which method is which?)
      
            Once each generation has finished, all data is pulled to local
            workstation in form of sqlite databases (1 database per job)
            and these are analysed and the fitness estimated sequentially
            the fitness array is then returned.
        """
        
        import time
        import ssh_utils
        
        self.CandidateData_list=[]
       
        self.__build_condor_files(candidates) #Build submit and runx.sh files, exp_id now corresponds to position in chromosome and fitness arrays
        
        messagehost=ssh_utils.host(optimizer_params.host,optimizer_params.username,optimizer_params.password,optimizer_params.port)
        
        self.__delete_remote_files__(messagehost)#delete everything in thssh_utilse directory you're about to put files in
        filelist=os.listdir(self.tmpdir)
        self.__put_multiple_files(messagehost,filelist,localdir=self.tmpdir,remotedir=optimizer_params.remotedir)#copy local files over
        filelist=os.listdir(self.portableswdir)
        self.__put_multiple_files(messagehost,filelist,localdir=self.portableswdir,remotedir=optimizer_params.remotedir)#copy local files over
        
        ssh_utils.issue_command(messagehost,'export PATH=/opt/Condor/release/bin:$PATH\ncondor_submit submitfile.submit')
        
        self.jobdbnames=[]
        for job_num in range(self.num_jobs):        #make a list of the databases we need:
            jobdbname='outputdb'+str(job_num)+'.sqlite'
            self.jobdbnames.append(jobdbname)
           
        #wait till you know file exists:
        dbs_created=False
        pulled_dbs=[] # list of databases which have been extracted from remote server
        while (dbs_created==False):
            print('waiting..')
            time.sleep(20)            
            print('checking if dbs created:')
            command='ls'
            remote_filelist=ssh_utils.issue_command(messagehost, command)

            for jobdbname in self.jobdbnames:

                db_exists=jobdbname+'\n' in remote_filelist
                
                if (db_exists==False):
                    print(jobdbname +' has not been generated')
                    dbs_created=False

                elif db_exists==True and jobdbname not in pulled_dbs:
                    print(jobdbname +' has been generated')
                    remotefile=optimizer_params.remotedir+jobdbname
                    localpath=os.path.join(self.datadir,str(self.generation)+jobdbname)
                    ssh_utils.get_file(messagehost,remotefile,localpath)
                    pulled_dbs.append(jobdbname) #so that it is not extracted more than once
                    #here pop-in the fitness evaluation

                if len(pulled_dbs)==len(self.jobdbnames):
                    dbs_created=True
            
        fitness=[]
        for CandidateData in self.CandidateData_list:
            job_num = CandidateData.job_num
            
            dbname=str(self.generation)+'outputdb'+str(job_num)+'.sqlite'
            dbpath=os.path.join(self.datadir,dbname)
            exp_id=CandidateData.exp_id
            
            connection=sqldbutils.db_connect(dbpath) #establish a database connection
            query='SELECT numerical_value\
                    FROM output_params WHERE experiment_id=\
                    '+str(exp_id)+' AND parameter="fitness"'       

            exp_fitness=sqldbutils.execute_query(connection,query)
            exp_fitness=exp_fitness.fetchall()
            exp_fitness=exp_fitness[0][0]
            
            #print('fitness: %s'%exp_fitness)
    
            fitness.append(exp_fitness)

        self.generation+=1
        return fitness

    def __local_evaluate(self,candidates,args):
        import time
     
        self.CandidateData_list=[]
        analysis_var=self.analysis_var
        
        #Build submitfile.submit and runx.sh files:
        self.__build_condor_files(candidates) #exp_id now corresponds to position in chromosome/fitness array
        
        fitness=[]
        #submit the jobs to the grid
        os.chdir(self.tmpdir)
        os.system('condor_submit submitfile.submit')
        
        #wait till you know file exists:
        dbs_created=False
        while (dbs_created==False):
            print('checking if dbs created:')
            for job_num in range(self.num_jobs):
                jobdbname='outputdb'+str(job_num)+'.sqlite'
                jobdbpath=os.path.join(self.datadir,jobdbname)
                print(jobdbpath)
                db_exists=os.path.exists(jobdbpath)
                
                if (db_exists==False):
                    time.sleep(60)
                    dbs_created=False
                    break
                
                dbs_created=True

        for CandidateData in self.CandidateData_list:
            job_num = CandidateData.job_num
            dbname='/outputdb'+str(job_num)+'.sqlite'
            dbpath=self.datadir+dbname
            exp_id=CandidateData.exp_id
            exp_data=sqldbutils.sim_data(dbpath,exp_id)
            analysis=analysis.IClampAnalysis(exp_data.samples,exp_data.t,analysis_var,5000,10000)
            exp_fitness=analysis.evaluate_fitness(optimizer_params.targets,optimizer_params.weights,cost_function=analysis.normalised_cost_function)
            fitness.append(exp_fitness)

        for job_num in range(self.num_jobs):
            jobdbname='outputdb'+str(job_num)+'.sqlite'
            jobdbpath=os.path.join(self.datadir,jobdbname)
            print(jobdbpath)
            os.remove(jobdbpath)

        return fitness
 '''


class PointBasedAnalysis(object):
    def __init__(self, v, t):

        self.v = numpy.array(v)
        self.t = numpy.array(t)

    def analyse(self, targets):
        analysis_results = {}

        for target in targets:
            target_time = float(target.split("_")[1])
            i = 0
            while self.t[i] < target_time:
                value = self.v[i]
                i += 1
            analysis_results[target] = value

        return analysis_results


class PointValueEvaluator(__Evaluator):
    """
    Locally-evaluates (not using cluster or grid computing) a model.

    The evaluate routine runs the model and returns its fitness value

    """

    def __init__(self, controller, parameters, weights, targets=None):

        super(PointValueEvaluator, self).__init__(
            parameters, weights, targets, controller
        )

    def evaluate(self, candidates, args):

        print("\n>>>>>  Evaluating: ")
        for cand in candidates:
            print(">>>>>       %s" % cand)

        simulations_data = self.controller.run(candidates, self.parameters)

        fitness = []

        for data in simulations_data:

            times = data[0]
            samples = data[1]

            data_analysis = PointBasedAnalysis(samples, times)

            fitness_value = self.evaluate_fitness(
                data_analysis, self.targets, self.weights
            )

            fitness.append(fitness_value)

            print("Fitness: %s\n" % fitness_value)

        return fitness

    def evaluate_fitness(
        self,
        data_analysis,
        target_dict={},
        target_weights=None,
        cost_function=normalised_cost_function,
    ):
        """
        Return the estimated fitness of the data, based on the cost function being used.

            :param data_analysis: PointBasedAnalysis instance
            :param target_dict: key-value pairs for targets
            :param target_weights: key-value pairs for target weights
            :param cost_function: cost function (callback) to assign individual targets sub-fitness.
        """

        fitness = 0

        analysed = data_analysis.analyse(target_dict)

        for target in target_dict.keys():

            target_value = target_dict[target]

            if target_weights is None:
                target_weight = 1
            else:
                if target in target_weights.keys():
                    target_weight = target_weights[target]
                else:
                    target_weight = 1.0

            if target_weight > 0:

                # let function pick Q automatically
                inc = target_weight * cost_function(analysed[target], target_value)
                fitness += inc

                print(
                    "Target %s (weight %s): target val: %s, actual: %s, fitness increment: %s"
                    % (target, target_weight, target_value, analysed[target], inc)
                )

        return fitness



"""

    Script to plot evolution of parameters in neurotune

"""

import math
import numpy as np
try:
    from typing import List, Dict, Bool, Union
except ImportError:
    pass

import logging

logger = logging.getLogger(__name__)


def plot_generation_evolution(
    sim_var_names,
    target_values={},
    individuals_file_name="../data/ga_individuals.csv",
    show_plot_already=True,
    save_to_file=False,
    save_to_file_scatter=False,
    save_to_file_hist=False,
    title_prefix="",
):
    # type: (List, Dict, str, Bool, Union[Bool, str], Union[Bool, str], Union[Bool, str], str) -> None
    """Plot generation evolution related graphs.

    :param sim_var_names: names of variables
    :type sim_var_names: list
    :param target_values: target values provided for fitting
    :type target_values: dict
    :param individuals_file_name: name of file storing data from individual generation runs
    :type individuals_file_name: str
    :param show_plot_already: whether showing plots should wait until plot() is called
    :type show_plot_already: bool
    :param save_to_file: name of file to save fitness plot to, False to not save
    :type save_to_file: str or bool
    :param save_to_file_scatter: name of file to save scatter plot to, False to not save
    :type save_to_file_scatter: str or bool
    :param save_to_file_hist: name of file to save histogram plot to, False to not save
    :type save_to_file_hist: str or bool
    :param title_prefix: prefix of plot title
    :type title_prefix: str

    """

    sim_var_names = list(sim_var_names)
    import matplotlib.pyplot as pyplot

    individuals_file = open(individuals_file_name)

    generations = []
    generations_all = []
    generations_offset = []

    f = []
    val_num = len(sim_var_names)
    nrows = math.ceil(math.sqrt(val_num))
    ncols = math.ceil(val_num / nrows)

    if val_num <= 3 or val_num == 5:
        nrows = val_num
        ncols = 1
    if val_num == 4:
        nrows = 2
        ncols = 2
    if val_num == 6:
        nrows = 3
        ncols = 2
    if val_num == 7 or val_num == 8:
        nrows = 4
        ncols = 2
    if val_num == 9:
        nrows = 3
        ncols = 3
    if val_num == 10:
        nrows = 5
        ncols = 2
    population_total = 0
    generations_total = 0

    for line in individuals_file:
        generation = int(line.split(",")[0])
        if generation == 0:
            population_total += 1
        generations_total = generation

    print(
        "Generating plots for %s variables over %s generations with population %s"
        % (val_num, generations_total, population_total)
    )
    logger.debug("Vals shown in %i rows x %i columns" % (nrows, ncols))

    vals = {}
    colours = {}
    sizes = {}
    ind_vals = {}

    for i in range(val_num):
        vals[i] = []
        colours[i] = []
        sizes[i] = []
        ind_vals[i] = {}

    # Go back to the starting of the file
    individuals_file.seek(0, 0)

    for line in individuals_file:
        main_info = line.split("[")[0]
        values = line.split("[")[1]
        generation = int(main_info.split(",")[0])
        individual = int(main_info.split(",")[1].strip())
        fitness = float(main_info.split(",")[2].strip())

        if individual == 0:
            logger.debug("Generation %s..." % generation)
            generations.append(generation)

        generations_all.append(generation)
        generations_offset.append(generation + (individual / 40.0))
        f.append(fitness)

        val_strings = values[:-2].split(",")

        for v in range(len(val_strings)):
            value = float(val_strings[v].strip())
            if individual == 0:
                ind_vals[v][generation] = []
            ind_vals[v][generation].append(value)
            vals[v].append(value)
            colours[v].append(individual)
            sizes[v].append((population_total - individual) * 2)

    fig1 = pyplot.figure()
    pyplot.get_current_fig_manager().set_window_title(
        title_prefix
        + " Evolution over %i generations of %s" % (generations_total, sim_var_names)
    )
    pyplot.subplots_adjust(hspace=0.4)
    pyplot.figtext(
        0.40,
        0.01,
        "Generation (%i individuals, offset slightly; larger circle => fitter)"
        % (population_total)
    )
    for i in range(val_num):

        var_name = sim_var_names[i]

        pyplot.subplot(nrows, ncols, i + 1, xlabel="Generation",
                       ylabel="{} ({})".format(var_name.split('/')[-2],
                                               var_name.split('/')[-1]))
        pyplot.title(var_name)
        if target_values is not None and sim_var_names[i] in target_values:
            value = target_values[sim_var_names[i]]
            x = [-1, generations_total + 1]
            y = [value, value]
            pyplot.plot(x, y, "--", color="grey")

        pyplot.scatter(generations_offset, vals[i], s=sizes[i], c=colours[i], alpha=0.4)

    if save_to_file_scatter:
        pyplot.savefig(save_to_file_scatter, dpi=300, bbox_inches='tight')

    fig2 = pyplot.figure()
    pyplot.get_current_fig_manager().set_window_title(
        title_prefix
        + " Fitness over %i generations from %s"
        % (generations_total, individuals_file_name)
    )

    ax = fig2.add_subplot(2, 1, 1, ylabel="Fitness")
    ax.scatter(generations_offset, f, s=sizes[i], c=colours[i], alpha=0.4)

    ax = fig2.add_subplot(2, 1, 2, ylabel="Fitness (log)")
    ax.set_yscale("log")
    ax.scatter(generations_offset, f, s=sizes[i], c=colours[i], alpha=0.4)
    pyplot.figtext(
        0.40,
        0.01,
        "Generation (%i individuals, offset slightly; larger circle => fitter)"
        % (population_total)
    )

    if save_to_file:
        pyplot.savefig(save_to_file, dpi=300, bbox_inches='tight')

    fig3 = pyplot.figure()
    pyplot.get_current_fig_manager().set_window_title(
        title_prefix + " Histograms over %i generations of %s" %
        (generations_total, sim_var_names)
    )

    pyplot.subplots_adjust(hspace=0.4)
    for i in range(val_num):

        var_name = (sim_var_names[i])
        ax = pyplot.subplot(nrows, ncols, i + 1,
                            xlabel="{} ({})".format(var_name.split('/')[-2],
                                                    var_name.split('/')[-1])
                            )
        pyplot.title(var_name)

        for generation in generations:
            values = ind_vals[i][generation]

            hist, bin_edges = np.histogram(values, bins=10)
            half_bin_width = (bin_edges[1] - bin_edges[0]) / 2
            xs = [be + half_bin_width for be in bin_edges[:-1]]

            shade = 1 - generation / (float(generations[-1]) + 1)
            # print("Gen: %s; shade: %s; value bins: %s; tots: %s"%(generation,shade,xs,hist))
            ax.plot(xs, hist, color=(shade, shade, shade))

    if save_to_file_hist:
        pyplot.savefig(save_to_file_hist, dpi=300, bbox_inches='tight')

    if show_plot_already:
        pyplot.show()

    individuals_file.close()


if __name__ == "__main__":

    # example 3
    target_values = {"amp": 65, "period": 250, "offset": -10}

    # example 2
    target_values = {
        "axon_gbar_na": 3661.79,
        "axon_gbar_kv": 23.23,
        "axon_gbar_kv3": 0.26,
        "soma_gbar_na": 79.91,
        "soma_gbar_kv": 0.58,
        "soma_gbar_kv3": 1.57,
    }

    parameters = [
        "leak_cond_density",
        "k_slow_cond_density",
        "k_fast_cond_density",
        "ca_boyle_cond_density",
        "specific_capacitance",
        "leak_erev",
        "k_slow_erev",
        "k_fast_erev",
        "ca_boyle_erev",
    ]

    # plot_generation_evolution(target_values.keys(), target_values)
    plot_generation_evolution(parameters)
